<?php

namespace App\Entity;

class SentenceNote extends Note
{
    const MODEL_NAME = 'Japanese sentences';

    private ?array $mediaInfo = [];
    private array  $terms = [];

    // -------------------------------------------------- Getters & setters ---

    public function getTerms(): array
    {
        return $this->terms;
    }
    public function setTerms(array $terms): static
    {
        $this->terms = $terms;
        return $this;
    }


    // ------------------------------------------------------- Anki-related ---

    public static function fromAnki(array $noteInfo): static
    {
        $note = parent::fromAnki($noteInfo);

        if ($note->getModel() !== self::MODEL_NAME) {
            throw new \Exception('Trying to parse wrong model');
        }

        $note->mediaInfo = self::parseMediaInfo($note->fields['Notes']);

        // Set VocabKanji field
        $note->terms = Term::fromNoteFields($note->fields);

        // If unable to, create them from the highlighted parts in the sentence
        if (empty($note->terms)) {
            foreach ($note->getHighlightedKanji() as $highlighedKanji) {
                $term = new Term();
                $term->kanji = $highlighedKanji;
                $term->definitionEn = null;
                $term->definitionJp = null;
                $note->terms[] = $term;
            }
        }

        // Set to null whatever is null
        $readings = array_map(
            fn($x) => in_array($x, ['_', '＿', '']) ? null : $x,
            explode('｜', $note->fields['VocabFurigana']),
        );

        // Set readings from furigana field
        foreach ($note->terms as $key => &$term) {
            if (null === $term->getReading()) {
                if (null !== ($readings[$key] ?? null)) {
                    $term->kanji .= '[' . $readings[$key] . ']';
                }
            }
        }

        return $note;
    }

    public function toAnki(): array
    {
        return array_merge(parent::toAnki(), [
            'fields' => [
                'VocabKanji' => join('｜', array_map(
                    fn(Term $x) => $x->getKanji(),
                    $this->terms,
                )),
                'VocabFurigana' => join('｜', array_map(
                    fn(Term $x) => $x->getReading() ?? '＿',
                    $this->terms,
                )),
                'VocabDef' => join("<br>\n", array_map(
                    fn(Term $x) => $x->toAnkiVocabDef(),
                    $this->terms,
                )),
            ],
        ]);
    }


    // ---------------------------------------------------- Derived methods ---

    public function hasTerm(string $kanji): bool
    {
        foreach ($this->terms as $term) {
            assert($term instanceof Term);

            if ($term->kanji == $kanji) return true;
        }

        return false;
    }

    public function isSentKanjiHighlighted(): bool
    {
        return str_contains(
            $this->fields['SentKanji'],
            self::HIGHLIGHT_ATTR_KANJI,
        );
    }

    /** Return an array of strings with the highlighted kanji in the SentKanji */
    public function getHighlightedKanji(): array
    {
        $ret = [];
        $matches = [];

        // 1. Get all spans in the text
        preg_match_all(
            self::HIGHLIGHT_PATTERN,
            $this->fields['SentKanji'],
            $matches,
            PREG_SET_ORDER,
        );

        // 2. Check the ones that match with the kanji color
        foreach ($matches as $match) {
            if ($match[1] === self::HIGHLIGHT_ATTR_KANJI) {
                $ret[] = mb_trim($match[2]);
            }
        }

        return $ret;
    }

    private static function parseMediaInfo(string $notes): ?array
    {
        $matches = null;

        // Parse the notes fields. It can be in the form of
        //   series-name_S01 EP07 (11h22m33s44ms)
        // or
        //   movie-name EP (11h22m33s44ms)
        if (1 !== preg_match(
            '/(?<name>[0-9A-Za-z\-_]+)(_S)?(?<season>\d*) EP(?<episode>\d*) \((?<time>.*)\)/n',
            $notes,
            $matches,
        )) {
            return null;
        }

        // Remove number-indexed matches, cast numbers to integers
        $matches = [
            'name' => $matches['name'],
            'time' => $matches['time'],
            // NOTE: intval returns 0 if not a number, which is false-like
            'season'  => intval($matches['season'])  ?: null,
            'episode' => intval($matches['episode']) ?: null,
        ];

        // Parse time into a DateInterval and replace it in the matches array
        $time = new \DateInterval('PT0S');

        preg_match('/(\d+)ms/', $matches['time'], $milliseconds);
        preg_match('/(\d+)s/',  $matches['time'], $seconds);
        preg_match('/(\d+)m/',  $matches['time'], $minutes);
        preg_match('/(\d+)h/',  $matches['time'], $hours);

        if ($milliseconds[1] ?? false) $time->f = $milliseconds[1] * 1000;
        if ($seconds[1]      ?? false) $time->s = $seconds[1];
        if ($minutes[1]      ?? false) $time->i = $minutes[1];
        if ($hours[1]        ?? false) $time->h = $hours[1];

        $matches['time'] = $time;

        return $matches;
    }
}
